
/*	Load / Save snapshots 

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.aug.94		started work (on application.c)			KIO !
	05.dec.94		replaced 'core' save&load by '.sna'		KIO !
	06.dec.94		added '.z80' v1.45 loader			KIO !
	10.dec.94		enhanced '.z80' loader to version 2.01		KIO !
	14.jan.95		separated this file from application.c		KIO !
	22.jan.95		save '.z80' added				KIO !
	05.nov.95		paged memory model support			KIO !
*/

#include	<string.h>
#include	<errno.h>
#include	<sys/stat.h>
#include	<sys/mman.h>

#include	"kio.h"
#include	"z80/z80.h"
#include	"h/options.h"
#include	"h/zxsp.h"
#include	"h/files.h"
#include	"h/main.h"
#include	"h/ay.h"
#include	"h/util.h"
#include	"h/screen.h"

#ifndef LOG
#define LOG 2
#endif
#ifndef SAFE
#define SAFE 3
#endif


// *****************************************************************


#if 0
/* ----	Compress .z80 file image -------------------------------------------
	compression sheme:  dc.b $ed, $ed, count, char
	return 0  --> ok --> compressed
	return !0 --> ok --> compressed, but compression increased size
	For security: the compression buffer must be at least  (5/3*sourcesize)  bytes long
*/
static OSErr CompressZ80 (Char *source, long sourcesize, Char *dest, long *destsize)
{	Char *q,*z,c,n,*qe;
	q = source;
	z = dest;
	qe = source+ sourcesize;
	
	XXXLogLine(">CompressZ80(...)");

	while (q<qe)
	{	c=*q++;
		if (*q!=c)					// single byte
		{	*z++ = c;
			if ((c==0xed)&&(q+2<=qe)&&(*q==*(q+1)))	// special care for compressible sequence after single 0xed !!!
				*z++=*q++;
			continue;
		}
		for (n=1; n<255 && *q==c; n++) q++;
		if (q>qe) { n -= q-qe; q=qe; };
		if ( (n>=4) || (c==0xed) )			// compress ?
		{	*z++ = 0xed;
			*z++ = 0xed;
			*z++ = n;
			*z++ = c;
		}
		else						// don't compress
		{	for (;n;n--) *z++ = c;
		}
	}
	
	*destsize = z-dest;

	XXXLogExit();

	return ( (*destsize<=sourcesize) ? ok : error );
}
#endif


/* ----	Decompress .z80 file image -----------------------------------------
	compression scheme:  dc.b $ed, $ed, count, char
	return 0:	ok
	return 1:	error
*/
static OSErr DecompressZ80 (Char *source, long sourcesize, Char *dest, long destsize)
{	Char *q,*z,c,n;

	XXXLogLine(">DecompressZ80(...)");

	q=source; source+=sourcesize;
	z=dest;	  dest+=destsize;
	
	while ( z<dest && q<source ) 
	{	c=*q++;						// read next byte
		if (c!=0xed) { *z++=c; continue; }		// single byte

		if (q==source) { *z++=c; continue; } 		// rare case: last source byte is 0xed

		c=*q++;						// read next byte
		if (c!=0xed) { *z++=0xed; q--; continue; };
		
		n=*q++; c=*q++;					// read count & char
		if ( (z+n)>dest ) goto x;
		if (q>source) goto x;
		MemSet (c, z, n);				// expand compressed bytes
		z+=n;
	};
	if (q==source&&z==dest) goto y;

	if (z==dest&&q+4<=source&&destsize==0xC000)	// version 1.45 file with end marker ?
	if (*q++==0) if (*q++==0xed) if (*q++==0xed) 
	if (*q++==0) 
	{
y:		XXXLog("ok");
		errno = ok;
	}
	else
	{
x:		XXXLog("error");
		errno = error;
	}

	XXXLogExit();

	return errno;
}


#if 0
/* ----	save one page in .z80 v2.01 format ---------------------------------------------------
	5.nov.95:	now takes pointer to page as argument
	3.jan.99:	linux port
*/
static OSErr save_block_z80 ( FILE* file, Char ID, Char* page )
{	Ptr		scratch;
	long	destsize;
	size_t 	n;
	Char	*p;
	
	XXXLogLine(">save_block_z80(...)");

	scratch = new Char[pagesize*5/3+4];
	if (!scratch) { Log("No space for compression buffer"); return outofmemory; };
	
	p = scratch;
	CompressZ80 ( page, pagesize, p+3, &destsize);
	*p++ = destsize;
	*p++ = destsize>>8;
	*p++ = ID;

	n = fwrite ( scratch, destsize+3, 1, file );
	delete[] scratch;
	if (n==0) Log("Error while writing RAM image: %s",strerror(ferror(file)));

	XXXLogExit();

 	return ferror(file);
}
#endif


struct snahead
{ 
	Char i,l2,h2,e2,d2,c2,b2,f2,a2,l,h,e,d,c,b,yl,yh,xl,xh,iff,r,f,a,spl,sph,im,brdr; 
};


static OSErr LoadSNA ( zxsp &specci, Ptr bu, size_t size )
{	OSErr err=error;

	XLogLine ( ">LoadSna(specci,%08X,%i)", bu, size ); 

	XTRAP ( pagesize!=0x4000 );
	
	snahead	&head 	= *(snahead*)bu;
	page	body	= bu + sizeof(snahead);
	Short	sp	= head.sph*256+head.spl;
	int	pages	= (size-sizeof(snahead))/pagesize;
	
	if ( (Short)(sp-1) <= (Short)0x4000 ) 
	{
		Log ( "Illegal stack pointer: $%04X", sp ); 
		goto x; 
	}

	if (head.im>2) 
	{ 
		Log ( "Illegal irpt mode: %i",head.im); 
		goto x; 
	}

	if ( sizeof(snahead)+pages*pagesize != size )
	{
		Log ( "Missaligned snapshot file size: %i (ignored)", size );
	}

	if (pages!=1&&pages!=3)
	{ 
		Log ( "Illegal number of pages in file: %i",pages ); 
		goto x; 
	}
	
// now do it

	specci.SetType(pages==3?zx48:zx16);

	for (int i=0;i<pages;i++)
		MemCopy ( body+pagesize*i, specci.GetRam(i+1), pagesize );

	specci.SetBorder ( head.brdr );

	specci.cpu.RI() = head.i;		
	specci.cpu.RR() = head.r;
	specci.cpu.RA() = head.a;		
	specci.cpu.RA2()= head.a2;		
	specci.cpu.RF()	= head.f;	
	specci.cpu.RF2()= head.f2;
	specci.cpu.RB()	= head.b;	
	specci.cpu.RC()	= head.c;
	specci.cpu.RD()	= head.d;	
	specci.cpu.RE()	= head.e;
	specci.cpu.RH()	= head.h;	
	specci.cpu.RL()	= head.l;
	specci.cpu.IX()	= head.xh*256+head.xl;	
	specci.cpu.IY()	= head.yh*256+head.yl;	
	specci.cpu.BC2()= head.b2*256+head.c2;	
	specci.cpu.DE2()= head.d2*256+head.e2;
	specci.cpu.HL2()= head.h2*256+head.l2;
	specci.cpu.IM()	= head.im;		
	specci.cpu.IFF1()=
	specci.cpu.IFF2()= (head.iff&0x04) ? enabled : disabled;
	specci.cpu.PCL()= specci.cpu.Peek(sp++);
	specci.cpu.PCH()= specci.cpu.Peek(sp++);	
	specci.cpu.SP()	= sp;

x:	XLogExit(); 
	return err;
}


/* ----	Load .scr screenshot file ----------------------------------------
*/
OSErr LoadSCR ( zxsp &specci, Ptr bu, size_t size )
{
	XLogLine(">LoadSCR(...)"); 

	if (size!=6912)
	{
		Log ( "wrong file length: %i", size );
		if (size<6912) { Log("aborting"); XLogExit(); return error; }
		Log("(ignored)");
	}

	MemCopy ( bu, specci.GetRam(1), 6912 );

	Char a,b,c,d; 
	a = bu[0x4000+6144+0] & 0x78;
	b = bu[0x4000+6144+31] & 0x78;
	c = bu[0x4000+6144+672] & 0x78;
	d = bu[0x4000+6144+703] & 0x78;

	if ( a==b || a==c || a==d ) {}
	else	if ( b==c || b==d ) a=b;
	else	if ( c==d ) a=c;

	specci.SetBorder ( (a>>3)&0x07 );

	XLogExit();

	return ok;
}


struct z80head
{
	// ----- Z80 Version 1.45 Header

	Char	a,f,c,b,l,h,pcl,pch,spl,sph,
		i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
		yl,yh,xl,xh,iff1,iff2,im,

	// ----- Z80 version 2.01 enhanced header (if PC=0)

		h2lenl,h2lenh,	// size of header extension: 23 for version 2.01 files
		npcl,npch,	// PC (instead of pcl+pch, which is set to 0)
		model,		// 	Value:	Meaning in v2.01	Meaning in v3.0
				//	0	48k			48k
				//	1	48k + If.1		48k + If.1
				//	2	SamRam			48k + M.G.T.
				//	3	128k			SamRam
				//	4	128k + If.1		128k
				//	5	-			128k + If.1
				//	6	-			128k + M.G.T.
		port_7ffd,	// If in SamRam mode, bitwise state of 74ls259.
				// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
				// If in 128 mode, contains last OUT to 7ffd (paging control)
		if1paged,	// !=0 means: interface1 rom is paged in
		rldiremu,	// Bit 0: 1 if R register emulation on
				// Bit 1: 1 if LDIR emulation on
		port_fffd,	// Last OUT to fffd (soundchip register number)
		soundreg[16],	// Contents of the sound chip registers

	// ----- Z80 version 3.0 enhanced header (if PC=0 & h2len=54)

		t_l,t_m,t_h,			// T state counter
		spectator,			// Flag byte used by Spectator (QL spec. emulator)
						// Ignored by Z80 when loading, zero when saving
		MGT_ROM,			// 0xFF if MGT Rom paged
		Multiface_ROM,			// 0fFF if Multiface Rom paged. Should always be 0.
		RAM0,RAM1,			// 0xFF if 0-8191 / 8192-16383 is RAM
		joy[10],			// 5* ascii word: keys for user defined joystick
		stick[10],			// 5* keyboard mappings corresponding to keys above
        	MGT_type,			// MGT type: 0=Disciple+Epson,1=Disciple+HP,16=Plus D
		disciple_inhibit_button_status,	// 0=out, 0ff=in
		disciple_inhibit_flag;		// Disciple inhibit flag: 0=rom pageable, 0ff=not
};

// length of headers:
#define z80v1len 30
#define z80v2len 55
#define z80v3len 86


/* ----	load .z80 file; version 1.45, 2.01, or version 3.0
*/
OSErr LoadZ80 ( zxsp &specci, Ptr bu, size_t size )
{	int loaded = 0;

	XLogLine(">LoadZ80(...)"); 
	XTRAP ( sizeof(z80head)!=z80v3len );

	{

// read registers:
	z80head head = *(z80head*)bu;
	if (head.data==0x0ff) head.data = 1;

// load version 1.45 file
	if (head.pch||head.pcl)
	{	
		bu += z80v1len;
		size -= z80v1len;
		
		specci.SetType(zx48);

		if (head.data&0x20)		// compressed
		{	
			XLog("Version 1.45 compressed"); 

			Ptr scratch;
			OMEM ( (scratch = new Char[3*pagesize]) );

			errno = DecompressZ80 ( bu, size, scratch, 3*pagesize );
			if (errno) 
			{ 	Log("The image is corrupted ... aborted!"); 
			}
			else
			{	MemCopy ( scratch+0*pagesize, specci.GetRam(1), pagesize );
				MemCopy ( scratch+1*pagesize, specci.GetRam(2), pagesize );
				MemCopy ( scratch+2*pagesize, specci.GetRam(3), pagesize );
			}
			delete[] scratch;
		}
		else	// uncompressed
		{
			XLog("Version 1.45 uncompressed");
			
			if ( size < 3*pagesize )
			{
				Log ( "image file is too short! ... aborted!" );
				errno=error;
			}
			else
			{	MemCopy ( bu+0*pagesize, specci.GetRam(1), pagesize );
				MemCopy ( bu+1*pagesize, specci.GetRam(2), pagesize );
				MemCopy ( bu+2*pagesize, specci.GetRam(3), pagesize );
			}
		};
		MemClear(head.soundreg,sizeof(head.soundreg));
		head.model=0;
		if (errno) goto x; else goto xok;
	};

// read extended header, at least for version 2.01
	switch ( head.h2lenl+256*head.h2lenh )
	{
	case 23:
		XLog("Version 2.01");	// version 2.01 header: ok
		break;
	case 54:
		XLog("Version 3.0");	// version 3.0 header: ok
		break;
	default:
		Log("Unrecognized version of .z80 file"); 
		errno = error; goto x;
	}

	head.pcl=head.npcl; head.pch=head.npch;
	bu += z80v1len+2+head.h2lenl;
	size-=z80v1len+2+head.h2lenl;

	if (head.if1paged) 
	{ 
		Log ( "Interface-1 ROM paged in (not supported)" );
		errno = error; goto x; 
	};		


// convert header from vs. 2.0 -> 3.0
	if (head.h2lenl==23)
	{
		XXXLog("Converting header from 2.01 -> 3.0");
		if (head.model>=2) head.model += 1;
		head.t_h = 0; 
		head.t_m = 0;
		head.t_l = 0;
		head.spectator=0;
		head.MGT_ROM = 0;
		head.Multiface_ROM = 0;
		head.RAM0 = head.RAM1 = 0;
		MemClear ( head.joy, 10 );
		MemClear ( head.stick, 10 );
		head.MGT_type = 0;
		head.disciple_inhibit_button_status = 0;
		head.disciple_inhibit_flag = 0;
		
		head.h2lenl = 54;
	}

// load version 3.0 file

// select required hardware
	switch (head.model)	
	{
	case 1:	Log ( "Warning: Interface-1 may be required!" );
	case 0:	XLog ( "Model = ZX Spectrum 48k" );
		specci.SetType(zx48);
		break;
	case 5:	Log ( "Warning: Interface-1 may be required!" );
	case 4:	XLog ( "Model = ZX Spectrum 128k" );
		specci.SetType(zx128);
		specci.PageCtrl(head.port_7ffd);
		break;
	case 6:
	case 2:	Log ( "M.G.T. interface required (not supported)");
		errno = error; goto x;
	case 3:	Log ( "SAMRAM required (not supported)");
		errno = error; goto x;
	default:Log ( "Unknown hardware configuration: &i", head.model );
		errno = error; goto x;
	}
		
// set T cycle counter
	specci.endcycles  -= specci.cpu.Cycles();
	specci.cpu.Cycles() = (Short)(head.t_m*256+head.t_l) + specci.FrameCycles()/4*(3-(head.t_h+1)%4) - specci.FrameCycles();
	specci.endcycles  += specci.cpu.Cycles();

// loop: load all pages
	errno = ok;
	while(size>3 && !errno)
	{
		struct bh { Char  lenl,lenh,nr; };
		int len;
		
// read page header
		bh& blockhead = *(bh*)bu; bu+=3;size-=3;
			
// read and decompress page
		len = blockhead.lenl+256*blockhead.lenh;
		if ( (int)size-(int)len<0 ) break;
		Ptr q;
		OMEM ( (q = new Char[len]) );
		MemCopy ( bu, q, len ); 
		bu+=len; size-=len;

		if (head.model<=2)	// zx spectrum 48k
		{	loaded |= 0xf1;
			switch (blockhead.nr)
			{
			case 0:	XXLog("ID 0: The ROM (skipped)");		break;
			case 1:	XXLog("ID 1: Interface-1 ROM (skipped)");	break;
			case 11:XXLog("ID 11: Multiface ROM (skipped)");	break;
			case 8:	XXLog("ID 8: RAM $4000-$7FFF");	
				errno = DecompressZ80 ( q,len, specci.GetRam(1),pagesize );	
				loaded |= 2; break;
			case 4:	XXLog("ID 4: RAM $8000-$BFFF");	
				errno = DecompressZ80 ( q,len, specci.GetRam(2),pagesize );	
				loaded |= 4; break;
			case 5:	XXLog("ID 5: RAM $C000-$FFFF");	
				errno = DecompressZ80 ( q,len, specci.GetRam(3),pagesize );	
				loaded |= 8; break;
			default:Log("Unknown page ID: %i", blockhead.nr); 
				errno=error; break;
			};
		}
		else			// zx 128k
		{	switch (blockhead.nr)
			{
			case 0:	XXLog("ID 0: Basic ROM (skipped)");		break;
			case 1:	XXLog("ID 1: Interface-1 ROM (skipped)");	break;
			case 2:	XXLog("ID 2: Boot ROM (skipped)");		break;
			case 11:XXLog("ID 11: Multiface ROM (skipped)");	break;
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:	
			case 8:
			case 9:
			case 10:Log ( "ID %i: RAM page %i", blockhead.nr, blockhead.nr-3 );
				errno = DecompressZ80 ( q, len, specci.GetRam(blockhead.nr-3), pagesize );
				loaded |= 1<<(blockhead.nr-3);
				break;
			default:Log("Unknown page ID: ", blockhead.nr); 
				errno=error; break;
			};
		}
			
		delete[] q;
	};


	if (errno) goto x;
	if (loaded!=0xff) { Log("Some RAM pages missing!"); specci.Reset(); errno = error; goto x; }
	if (size!=0) Log("Some extra bytes at end of file (ignored)");
	
// ----	store registers and resume execution
xok:	specci.cpu.RI()  = head.i;
	specci.cpu.RR()  = head.r&0x7f + head.data<<7;	// bit 7 von r steckt in data bit 0 ... tzz tzz
	specci.cpu.RA()  = head.a;		
	specci.cpu.RA2() = head.a2;		
	specci.cpu.RF()	 = head.f;		
	specci.cpu.RF2() = head.f2;
	specci.cpu.RB()	 = head.b;		
	specci.cpu.RC()  = head.c;	
	specci.cpu.RD()	 = head.d;		
	specci.cpu.RE()	 = head.e;
	specci.cpu.RH()	 = head.h;		
	specci.cpu.RL()	 = head.l;
	specci.cpu.IX()	 = head.xh *256 + head.xl;
	specci.cpu.IY()	 = head.yh *256 + head.yl;
	specci.cpu.SP()	 = head.sph*256 + head.spl;	
	specci.cpu.PC()	 = head.pch*256 + head.pcl;
	specci.cpu.BC2() = head.b2 *256 + head.c2;	
	specci.cpu.DE2() = head.d2 *256 + head.e2;
	specci.cpu.HL2() = head.h2 *256 + head.l2;
	specci.cpu.IFF1()= head.iff1;
	specci.cpu.IFF2()= head.iff2;

	specci.SetBorder ( head.data>>1 );

	specci.cpu.IM()	 = head.im&3;	
	if ((head.im&3)>2) { Log("Illegal interrupt mode: 3 (fixed: set to 1)");  specci.cpu.IM()=1; }

// if zx48: test, whether ay regs are supplied:  ((all registers 0x00 or all registers 0xff))
	Char m1=0,m2=0xff;
	for (int i=0;i<=15;i++) 
	{
		m1|=head.soundreg[i];
		m2&=head.soundreg[i];
	}

// init soundchip:
	m1 = (head.model>2) || (m1!=0 && m2!=0xff);	// bool: enabled ?
	specci.soundchip.EnableAY(m1);
	if ( m1 )
	{
		XLog("soundchip data supplied");
		for (int i=0;i<=15;i++)	
		{
			specci.soundchip.SetRegister(i, head.soundreg[i]);
		}
		specci.soundchip.SelectRegister(head.port_fffd);
	}

	errno = ok;	
	}

x:	XLogExit();
	return errno;
}


/* ----	Load ROM from file -------------------------------------------------
	Load ROM from file and reset Specci
	return ok	-> you should rename specci window
	else		-> you should reload roms and reset specci

OSErr LoadROM ( short fileID )
{
	PrintMsg(CStr("Load .ROM file")); 

	errno = File2Buffer ( fileID, rom[0], 0x4000 );
	if (errno)	PrintNumberAlert(CStr(localized("Error while reading file:","Fehler beim Lesen der Datei:")),errno);
	else		HandleReset(false);

	romID = errno?0:fileID;
	return errno;
}
*/


OSErr LoadFile ( zxsp &specci, char *name )
{
	XXLogLine ( ">LoadFile(specci,\"%s\")", name );

	size_t	size=0;
	Ptr	bu = (Ptr)-1;
	FILE *fp;
	struct stat sbuf;
	errno=noerror;
	
	if ( stat(name, &sbuf) == -1 ) { Log("stat() didn't work"); goto y; }
	if (!S_ISREG(sbuf.st_mode)) { Log ( "not a regular file" ); goto y; }
	size = (size_t)sbuf.st_size;
	if ( !(fp = Fopen(name, "rb")) ) { Log ( "couldn't open \"%s\" for read" ); goto y; }
	bu = (Ptr)mmap ( NULL, size, PROT_READ, MAP_SHARED, fileno(fp), (off_t)0 );
	fclose(fp);
	if ((long)bu == -1) { Log ( "mmap failed" ); y: errno=errno?errno:error; goto x; }

	if 
	(	strcmp(".sna",name+(strlen(name)-4))==0 ||
		strcmp(".SNA",name+(strlen(name)-4))==0
	)
		errno = LoadSNA ( specci, bu, size );
	else if
	(	strcmp(".z80",name+(strlen(name)-4))==0 ||
		strcmp(".Z80",name+(strlen(name)-4))==0
	)
		errno = LoadZ80 ( specci, bu, size );
	else if
	(	strcmp(".scr",name+(strlen(name)-4))==0 ||
		strcmp(".SCR",name+(strlen(name)-4))==0
	)
		errno = LoadSCR ( specci, bu, size );
	else
	{	Log ( "can't determine file type of %s", name );
		errno=error;
	}	
	
x:	if((long)bu!=-1) munmap((char*)bu, size);
	XXLogExit();	
	return errno;
}


/* ----	Save Core as .sna file ------------------------------------------------------------
	note: .sna files were originally saved from an NMI routine. 
	therefore the PC is stored on the Z80 stack and only IFF1 needs to be saved

OSErr SaveSNA ( short fileID )	
{	OSErr 	err;
	struct { Char i,l2,h2,e2,d2,c2,b2,f2,a2,l,h,e,d,c,b,yl,yh,xl,xh,iff,r,f,a,spl,sph,im,brdr; } head;
	
	PrintMsg(CStr("Save .sna snapshot file")); 

	if (z80_hardware!=zx48) PrintAlert ( CStr(localized(
		"Warning: The saved .sna snapshot contains only the 48 kbyte RAM of the ZX Spectrum 48k",
		"Achtung: Der gesicherte .sna Schnappschu enthlt nur die Daten fr einen ZX Spectrum 48k"  )));

	Poke(--SP,PCH);		// push retn-address
	Poke(--SP,PCL);
	
	head.im	= IM;	head.iff	= IFF1 ? 0x04 : 0x00;
	head.i	= RI;		head.r	= RR;
	head.a	= RA;	head.a2	= RA2;
	head.f	= RF;	head.f2	= RF2;
	head.b	= RB;	head.c	= RC;
	head.d	= RD;	head.e	= RE;
	head.h	= RH;	head.l	= RL;
	head.xh	= XH;	head.xl	= XL;
	head.yh	= YH;	head.yl	= YL;
	head.b2	= BC2>>8;	head.c2	= BC2;
	head.d2	= DE2>>8;	head.e2	= DE2;
	head.h2	= HL2>>8;	head.l2	= HL2;
	head.sph	= SPH;	head.spl	= SPL;
	head.brdr	= bordercolor;

	err = Buffer2File ( &head, fileID, 27 );
	if (err) { PrintNumberAlert(CStr(localized("Error while writing registers:","Fehler beim Schreiben der Register:")),err); return err; }

	err = Core2File ( 0x4000, fileID, 0xc000 );	
	if (err) PrintNumberAlert(CStr(localized("Error while writing RAM image:","Fehler beim Schreiben des Speicherauszuges:")),err);
	return err;
}
*/


/* ----	Save ROM to file -------------------------------------------------

OSErr SaveROM(short fileID)
{	OSErr err;

	PrintMsg(CStr("Save .ROM file")); 

	err = Buffer2File ( rom[0], fileID, 0x4000 );
	if (err) PrintNumberAlert(CStr(localized("Error while writing file:","Fehler beim Schreiben der Datei:")),err);
	return err;
}
*/


/* ----	Save .scr screenshot file ----------------------------------------

OSErr SaveSCR(short fileID)
{	OSErr err;

	PrintMsg(CStr("Save .SCR screenshot")); 

	err = Buffer2File ( video_page, fileID, 6912 );
	if (err) PrintNumberAlert(CStr(localized("Error writing file:","Fehler beim Schreiben der Datei:")),err);
	return err;
}
*/


#if 0
/* ----	save .z80 file; version 2.01 --------------------------------------------------
*/
OSErr SaveZ80 ( short fileID )
{	OSErr err;
	int i;

	struct {	Char 					// Z80 Version 1.45 Header
			a,f,c,b,l,h,pcl,pch,spl,sph,
			i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
			yl,yh,xl,xh,iff1,iff2,im,
									// Z80 Version 2.01 enhanced header (if PC=0)
			h2lenl,h2lenh,				// size of header extension: 23 for version 2.01 files
			npcl,npch,					// PC (instead of pcl+pch, which is set to 0)
			model,					// 	Value:	Meaning in v2.01	Meaning in v3.0
									//	0		48k				48k
									//	1		48k + If.1			48k + If.1
									//	2		SamRam			48k + M.G.T.
									//	3		128k				SamRam
									//	4		128k + If.1		128k
									//	5		-				128k + If.1
									//	6		-				128k + M.G.T.
			port_7ffd,				// If in SamRam mode, bitwise state of 74ls259.
									// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
									// If in 128 mode, contains last OUT to 7ffd (paging control)
			if1paged,					// !=0 means: interface1 rom is paged in
			rldiremu,					// Bit 0: 1 if R register emulation on
									// Bit 1: 1 if LDIR emulation on
			port_fffd,					// Last OUT to fffd (soundchip register number)
			soundreg[16];				// Contents of the sound chip registers
		} head;

	PrintMsg(CStr("Save .z80 snapshot file")); 

	MemClear ( &head, sizeof(head) );			// head.pcl = head.pch = head.if1paged = 0;

	head.h2lenl	= 23;	head.h2lenh = 0;	// flag: version 2.01 header
	
	switch(z80_hardware)
	{	case zx48:	head.model = 0;	 break;	// ZX Spectrum 48k
		case zx128:	head.model = 3; break;	// ZX 128
	}

	head.port_fffd	= AYregnr;				// ZX 128 only: value of AY register select register
	MemCopy ( AYreg, head.soundreg, 16 );		// ZX 128 only: AY registers
	head.port_7ffd	= zx128pagectrl;			// ZX 128 only: value of page control register

	head.im	= IM;
	head.i	= RI;
	head.r	= RR & 0x7F;
	head.a	= RA;
	head.a2	= RA2;
	head.f	= RF;
	head.f2	= RF2;
	head.b	= RB;		head.c	= RC;
	head.d	= RD;		head.e	= RE;
	head.h	= RH;		head.l	= RL;
	head.xh	= XH;		head.xl	= XL;
	head.yh	= YH;		head.yl	= YL;
	head.b2	= BC2>>8;		head.c2	= BC2;
	head.d2	= DE2>>8;		head.e2	= DE2;
	head.h2	= HL2>>8;		head.l2	= HL2;
	head.sph	= SPH;		head.spl	= SPL;
	head.npcl	= PCL;		head.npch	= PCH;

	head.iff1 = IFF1;
	head.iff2 = IFF2;
	
	head.data = ((RR&0x80)>>7) + bordercolor<<1;

	err = Buffer2File(&head,fileID,55);
	if (err) { PrintNumberAlert(CStr(localized("Error while writing registers:","Fehler beim Schreiben der Register:")),err); return err; }


/*	Now the compressed blocks follow. Each block has a 3 byte header:
		dc.w		length of data (without this header; low byte first)
		dc.b		page number of block
		dc.s		compressed data follows

	The pages are numbered, depending on the hardware mode, in the following way:

		Page		48 mode			128 mode			SamRam mode
		0		48K rom			rom (basic)		48K rom
		1		Interface I, Disciple or Plus D rom, according to setting
		2		-				rom (reset)		samram rom (basic)
		3		-				page 0			samram rom (monitor,..)
		4		8000-bfff			page 1			Normal 8000-bfff
		5		c000-ffff			page 2			Normal c000-ffff
		6		-				page 3			Shadow 8000-bfff
		7		-				page 4			Shadow c000-ffff
		8		4000-7fff			page 5			4000-7fff
		9		-				page 6			-
		10		-				page 7			-
		11		Multiface rom		Multiface rom		-

	In 48K mode, pages 4,5 and 8 are saved.
	In SamRam mode, pages 4 to 8 are saved.  
	In '128 mode, all pages from 3 to 10 are saved.

	The 128 has a memory map like:   Rom [switchable];   Ram 5;   Ram 2;   Ram [switchable]
	The paging is all controlled by OUT commands to 7FFDh (32765). This port is 
	write-only so a copy of the last byte sent to it is stored at 5B5Ch (23388).
	
	The byte to output is:
		Bits 0-2:	Ram for top 16k:	0 is the usual bank;   7 contains some scratchpads used by the 128k editor.
		Bit 3:	Screen select:		0 => Use the normal screen in bank 5;	1 => Use the shadow screen in bank 7.
		Bit 4:	ROM for lower 16k:	0 => 128k editor;					1 => 48k BASIC
		Bit 5:	Lock. If set, disables all further paging.
*/

	if (z80_hardware==zx48)
	{			err = save_block_z80 ( fileID, 4, RdPtr(0x8000) );
		if (!err)	err = save_block_z80 ( fileID, 5, RdPtr(0xC000) );
		if (!err)	err = save_block_z80 ( fileID, 8, RdPtr(0x4000) );
	}

	if (z80_hardware==zx128)
	{	for (i=0;i<=7&&!err;i++)   err = save_block_z80 ( fileID, i+3, ram[i] );
	}

	if (err) PrintNumberAlert(CStr(localized("Error while writing core:","Fehler beim Schreiben der Speicherabzuges:")),err);

	return err;
}
#endif


