/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#ifndef audio_h
#define	audio_h	audio_h


/* ====================================================
		audio output
		hardware dependent part
===================================================== */


// ----	adjustable values for dsp ----------------------
#define DEV_AUDIO	"/dev/dsp"		// audio device
#define	DEV_AUDIO_AUX				// none
#define	SAMPLECLOCK	35000			// preferred sample clock
#define	CHUNKS		12			// buffer size of dsp: approx. space for 0.1 sec
#define	CHUNKBITS	9
#define	CHUNKSIZE	(1<<CHUNKBITS)		// size of chunk (samples, not bytes!)


// ----	type for one single sample ---------------------
typedef	Char		Sample;			// e.g.: stereo 16bit -> 4 bytes!
const	Sample		quiet	= 0x80;		// center value for sound samples


/* ----	the digital sample player ----------------------
	the hardware which actually plays the sound
*/
class Dsp
{
	int	fd;				// handle to audio device
	Sample	*buffer;			// start of sound buffer
	Sample	*bufferend;			// end of sound buffer
	Sample	*bufferptr;			// write pointer

	int	clock;				// sample clock of dsp

	OSErr	Write	( );			// actually write buffer

public:
		Dsp	( );
		~Dsp	( );

	int	Clock	( )			{ return clock; }
	int	State	( );
	bool	On	( )			{ return fd!=0; }

	OSErr	Open	( );			// open sound device
	void	Close	( );			// close & release device
	OSErr	Write	( Sample*, int count );	// write block of samples
inline	void	Write	( Sample s );		// write one single sample
};


// ----	values for state ----------------------------
enum 
{ 	dsp_off=0, 	// no sound: disabled
	dsp_busy,	// no sound: device busy
	dsp_empty,	// sound on: buffer empty		=> hurry up!
	dsp_low,	// sound on: buffer usage < 30%		=> hurry up!
	dsp_fine,	// sound on: no need to hurry up or wait
	dsp_high,	// sound on: buffer usage > 70%		=> slow down, writing may block!
	dsp_full,	// sound on: buffer full		=> slow down, writing will block!
};


/* ----	send 1 sample to dsp ----------------------------------------
*/
inline void Dsp::Write ( Sample s )
{
	if ( bufferptr<bufferend || Write()==ok ) *bufferptr++ = s;
}


#endif










