/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef		zxsp_screen_h
#define		zxsp_screen_h zxsp_screen_h


#include	<X11/Xlib.h>
#include	"kio.h"
#include	"h/window.h"


#define	SCR_COLORS	 16
#define	SCR_HSIZE	(8*32)		// dimension of specci screen
#define	SCR_VSIZE	(8*24)		// ""
#define SCR_MAXSCALE	 4		// current max. supported magnification
#define	SCR_ABOVE	 48		// max. border lines above screen (max. 63)
#define	SCR_BELOW	 48		// max. border lines below screen (max. 56)

#define	SCR_BORDER	(SCR_LINES+SCR_ABOVE+SCR_BELOW)
#define	SCR_LINES	SCR_VSIZE	// for convenience...


typedef	Char	scrline[32];		// specci screen raster line


class Scr
{
	Win	window;			// X window

	long	color[SCR_COLORS];	// the 16 specci colors
	long	pap[256];		// convert attr -> paper color
	long	ink[256];		// convert attr -> ink color

	long	nextflash;		// frame number of next attribute flash
	bool 	flashphase;		// which phase ?

	long	dirty [SCR_LINES];	// dirty flags: 1 bit per screen byte
	scrline	pixels[SCR_LINES];	// screen contents at scanline output
	scrline	attr  [SCR_LINES];	// attributes at scanline output
	char	border[SCR_BORDER];	// border color at scanline output

	int	scale;			// current magnification
	int	xa,ya,xe,ye;		// current left and top border
	int	width,height;		// current window size in pixels
	int	method;			// id of refresh routine to use

	void	Refresh1	( );
	void	Refresh2	( );
	void	Refresh3	( );
	void	Refresh4	( );
	void	Refresh6	( );
	void	Refresh8	( );
	void	Refresh12	( );
	void	Refresh16	( );
	void	Refresh24	( );
	void	Refresh32	( );
	void	Refresh48	( );
	void	Refresh64	( );
	void	Refresh96	( );
	void	Refresh128	( );

	void	Refresh24x1	( );
	void	Refresh24x2	( );
	void	Refresh24x3	( );
	void	Refresh24x4	( );

	void	HLine		( long ink, int x, int y, int width );

public:
		Scr		( int& argc, char* argv[] );
		~Scr		( );

	void 	Sync		( );	// actually redraw screen now!

inline	void	UpdateScreen	( int line, scrline pixels, scrline attr );	// line number 0 = first line of screen
inline	void	UpdateBorder	( int line, Char newborder );			// line number 0 = first line of screen; top border line numbers are negative !!!
inline	int	LinesAbove	( )	{ return Min(ya,SCR_ABOVE); }
inline	int	LinesBelow	( )	{ return Min(height-ye,SCR_BELOW); }

	bool	PollEvents	( XEvent& xev );	// in: sys/windows.cc
};


inline void Scr::UpdateBorder ( int line, Char newborder )
{ 
	line += SCR_ABOVE;
	if ((Int)line<SCR_BORDER)
	{
		if ((newborder^border[line])&7) border[line] = newborder|0xF8;
	}
}


inline void Scr::UpdateScreen ( int line, scrline newpixels, scrline newattr )
{
	if ((Int)line<SCR_LINES)
	{
		long n=0;
		int i;
		Char *p1,*p2;
			
		p1 = newpixels;
		p2 = pixels[line];
		for (i=32;i--;) n += n+(*p1++!=*p2++);
		if (n) for (i=8;i--;) *--(long*&)p2 = *--(long*&)p1;
		dirty[line] |= n;

		p1 = newattr;
		p2 = attr[line];
		for (i=32;i--;) n += n + (*p1++!=*p2++);
		if (n) for (i=8;i--;) *--(long*&)p2 = *--(long*&)p1;
		dirty[line] |= n;
	}
}


#endif









