/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef 	ZXSP_H
#define	ZXSP_H ZXSP_H


#include	"kio.h"
#include	"z80.h"
#include	"ay.h"
#include	"keyboard.h"
#include	"screen.h"

#define	max_lines_above		topborderlines
#define	screen_lines		( 24*8 )
#define	max_lines_below		(framecycles/linecycles-topborderlines-screen_lines)


typedef	Char	*page;
const 	int 	maxrompages	= 4;
const	int	maxrampages	= 8;

class zxsp : micro
{

friend void 	Test	( );

	page	rom[maxrompages];	// rom0, rom1, if1, ext.
	page	ram[maxrampages];
	page	videopage;
	int	type;			// zx16, zx48, zx128, ...
	Char	rombits;
	Char	pagectrl;		// state of page register	(w)
	Char	border;			// state of border color	(w)
	Char	speaker;		// state of speaker output	(w)

	long	cpuclock;		// the crystal for the cpu

	long	framerate;		// ula: frames / second
	long	linecycles;		// ula: cpu clock cycles / scan line
	long	topborderlines;		// ula: lines before writable screen starts

	long	framecycles;		// derived value: cpu clock cycles / frame

public:
	z80	cpu;			// z80 cpu
	Scr	screen;			// the video output circuitry
	kbd	keyboard;		// keyboard & joystick	
	ay	soundchip;		// soundchip and speaker

public:
	long	nextframe;	// next frame to do. compare against global frameCounter which is incremented by interrupt
	long	endcycles;	// T cycle within frame when cpu.cycles will expire again
	long	scanline;	// current scanline

		zxsp		( int& argc, char* argv[] );
		~zxsp		( );

	void	PollEvents	( );	// poll X windows events  ((in sys/window.cc))

	void	SetType		( int type );
	void	PageCtrl	( Char iobyte );
	void	Reset		( );
	void	NMI		( );	
	void	Irpt		( Char busbyte=255 );
	void	Run		( );
	void	RST0		( );

	void	AttachRom	( int n, char* filename );


// querries:
	int	GetType		( )	{ return type; }
	page	Videopage	( )	{ return videopage; }
	page	GetRam		(int i)	{ return ram[i%maxrampages]; }
	page	GetRom		(int i)	{ return rom[i%maxrompages]; }
	long	Now		( )	{ return endcycles-cpu.Cycles(); }
	long	FrameCycles	( )	{ return framecycles; }


// write access:
	void	SetBorder	(int n)	{ border = n&7; }

// virtual overloaded functions
virtual	Char	Input		( long cc, Short addr );
virtual	void	Output		( long cc, Short addr, Char byte );
virtual	long	Cycles		( long cc );

virtual void	Info_Misc	( long cc, Short ip );		// info call for interesting instruction
virtual void	Info_Illegal	( long cc, Short ip );		// info call for illegal instruction
virtual void	Info_Weird	( long cc, Short ip );		// info call for weird illegal instruction
virtual void	Info_Irpt	( long cc, Short ip );		// info call if INFO_IRPT==on
virtual void	Info_NMI	( long cc, Short ip );		// info call if INFO_NMI==on
virtual void	Info_1st_Loc	( long cc, Short ip );		// info call if PC_PROFILE==on
virtual void	Info_1st_Instr	( long cc, Short ip );		// info call if CMD_PROFILE==on
};


enum	{ zx16,zx48,zx128,plus2,plus3 };




#endif



