
/*	This program emulates a ZX Spectrum computer.

***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
			
	01.Aug.94	started work on Macintosh version	kio
	24.Dec.98	started port to linux				kio
*/

// ***** The Main Application File ******************************

#include	<unistd.h>
#include	<signal.h>
#include	<string.h>
#include	<time.h>
#include	<sys/time.h>
 

#include	"h/options.h"

#ifndef		LOG
#define		LOG 1
#endif
#ifndef		SAFE
#define		SAFE 2
#endif

#include	"z80/z80.h"
#include	"h/main.h"
#include 	"h/util.h"
#include 	"h/screen.h"
#include 	"h/main.h"
#include	"h/files.h"
#include	"h/ay.h"


/* ----	global data ----------------------------------------------
*/

const char	*applDate	= __DATE__;
const char	*applTime	= __TIME__;
const char	*applName	= "zxsp";
const char	*applVersion= "0.4.2";

const char	*myCopyright= "Copyright (c) 1994-2000 Gnter Woigk";
const char	*myName		= "Gnter Woigk";
const char	*myAddress	= "Sankt Johann 6/246, 91056 Erlangen, Germany";
const char	*myEmail	= "kio@myokay.net";


OSErr		err		= noerror;
bool		done	= false;
bool		abo		= false;

char		*progName;		// supplied by main()
time_t	 	startupTime;	// supplied by main()
int 		stdinFileType;	// supplied by main()
int			stdoutFileType;	// supplied by main()

long		frameCounter=0;	// global frame counter


char*		errorText[]	=
{
	"General error",
	"Internal error"
};

int		errorTextSize	= NELEM(errorText);



/* **************************************************************** */
/* ***********    A  P  P  L  I  K  A  T  I  O  N    ************** */
/* **************************************************************** */


/* ----	Copyright ------------------------------------------------
*/
void PrintCopyright ( )
{
	LogLine ( "zxsp-x for X11@Linux Version  %s", applVersion );
	LogLine ( "Compiled: %s at %s", applDate, applTime  );
	LogLine ( "%s \n", myCopyright );
	
	LogLine ( "%s; %s", myName, myAddress );
	LogLine ( "mailto:%s \n", myEmail );
}


/* ----------------------------------------------------------
	the system timer
---------------------------------------------------------- */


void IntFrequency ( int msec )
{
	XXLogLine ( ">IntFrequency(%i)", msec );
	
	struct itimerval itv;
	itv.it_interval.tv_sec	= 0;
	itv.it_interval.tv_usec	= msec*1000;
	itv.it_value.tv_sec	= 0;
	itv.it_value.tv_usec	= msec*1000;
	TRAP ( setitimer(ITIMER_REAL, &itv, NULL) );

	XXLogExit();
}


static void CatchSignals(int signo)
{
	if (signo==SIGALRM) 
	{
		if (XXXLOG) { timeval tv; gettimeofday(&tv,0); LogLine ( "CatchSignal(SIGALRM) --> %i", tv.tv_usec ); }
		frameCounter++;
		return;
	}

	if (signo == SIGSEGV)
	{
		static int core = 0;
		if (core++) exit(3);
		LogLine ( "Signal handler caught signal SIGSEGV" );
	}
	Quit(2);
}


static void InitSignals ( bool onoff )
{	struct sigaction sigact;

	sigemptyset ( &sigact.sa_mask );

	#ifdef SA_RESTART
	sigact.sa_flags = SA_RESTART;
	#else
	sigact.sa_flags = 0;
	#endif

// signals always ignored:
	sigact.sa_handler = SIG_IGN;

	#ifdef SIG_XCPU
	sigaction(SIG_XCPU, &sigact, NULL);
	#endif

// signals ignored or handled:	
	if (onoff) sigact.sa_handler = CatchSignals;
	sigaction(SIGINT,   &sigact, NULL);
	sigaction(SIGHUP,   &sigact, NULL);
	sigaction(SIGTERM,  &sigact, NULL);
	sigaction(SIGQUIT,  &sigact, NULL);
	sigaction(SIGALRM,  &sigact, NULL);

// signals always handled:	
	sigact.sa_handler = CatchSignals;
	sigaction(SIGSEGV,  &sigact, NULL);
}


int main ( int argc, char**argv )
{	char *s;

	XLogLine ( ">main(...)" );

	for (int i=0;i<argc;i++) XLog ( "%s", argv[i] );

	startupTime = time(NULL);

	if ( (s = strrchr(*argv, '/') ) != NULL ) progName = ++s;
	else	progName = *argv;
	argv++; argc--;
	XXLogLine ( "my name: %s", progName );

	PrintCopyright();
	InitSignals(off);			// prevent termination while initializing

	zxsp specci(argc,argv);		// this is our specci instance !

	InitSignals(on);
	IntFrequency(20);

	specci.SetType(zx48);
	specci.soundchip.Resume(0);

	if ( argc ) 
	{
		XXLogLine ( "unused arguments:" );
		for (int i=0;i<argc;i++) XXLog("--> %s",argv[i]);
		LoadFile(specci,argv[0]);
	}

	while (!done) specci.Run();

	Quit(0);
	XLogExit();
	XLogLine("");
	return 0;
}









