/*
***	Copyright:	1998-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "h/options.h"
#ifndef LOG
#define LOG 1
#endif
#ifndef SAFE
#define SAFE 2
#endif

#include "options.h"
#include "screen.h"
#include "util.h"
#include "main.h"
#include "keyboard.h"
//#include "debug.h"
//#include "zxsp.icon"
#include "h/window.h"



/* ------------------------------------------------------------
			screen refresh
			hardware independent part
------------------------------------------------------------ */


Scr::Scr ( int& argc, char* argv[] )
:	window(argc,argv)
{
	if (done) return;	// s.th. utterly went wrong ...

	XXLogLine ( ">Scr::Scr(%i,argv)", argc );

// initialize X window ink table
	for (uint i=0;i<NELEM(color);i++)  
	{
		color[i] = window.Color(i);
	}

// initialize color conversion tables
	for ( int i=0; i<256; i++ )
	{
		ink[i]	=  color[(i&7)+((i>>3)&8)];		// convert attribute byte -> color index of ink
		pap[i]	=  color[(i>>3)&0xf];			// convert attribute byte -> color index of paper

	//	bwm[i]	=  (i&7) == ((i>>3)&7)  ?  0 : 0xff;	// mask to ignore pixel bits if paper == ink
	//	bwn[i] 	=  bwm[i]  ?  (  (i&7) < ((i>>3)&7)  ?  0 : 0xff  )  :  (  (i&7) > 3 ? 0 : 0xff  );	// mask to invert pixel bits
	};

// init flashing:
	nextflash = frameCounter;	// frameCounter is incremented by interrupt
	flashphase = 1;

// init screen and border
	MemClear(attr,  sizeof(attr));		// all attributes := black on black
	MemClear(pixels,sizeof(pixels));	// all screen pixels := '0'
	MemSet  (0xff,dirty,sizeof(dirty));	// offscrren image := not up to date
	MemSet  (0xf8,border,sizeof(border));	// entire border  := black & not up to date

	XXLogExit();
}


/* ----	destructor -----------------------------------------------------
*/
Scr::~Scr ( )
{
}


/* ======================================================================
			Now, refresh routines for
			different color depths and 
			screen magnification follow:
======================================================================= */

void Scr::Refresh1()		// 1bpp*1
#define SIZE 1
#include "refreshxy.c"

void Scr::Refresh2()		// 1bpp*2, 2bpp*1
#define SIZE 2
#include "refreshxy.c"

void Scr::Refresh3()		// 1bpp*3
#define SIZE 3
#include "refreshxy.c"

void Scr::Refresh4()		// 1bpp*4, 2bpp*2, 4bpp*1
#define SIZE 4
#include "refreshxy.c"

void Scr::Refresh6()		// 2bpp*3
#define SIZE 6
#include "refreshxy.c"

void Scr::Refresh8()		// 2bpp*4, 4bpp*2, 8bpp*1
#define SIZE 8
#include "refreshxy.c"

void Scr::Refresh12()		// 4bpp*3
#define SIZE 12
#include "refreshxy.c"

void Scr::Refresh16()		// 4bpp*4, 8bpp*2, 16bpp*1
#define SIZE 16
#include "refreshxy.c"

void Scr::Refresh24()		// 8bpp*3
#define SIZE 24
#include "refreshxy.c"

void Scr::Refresh32()		// 8bpp*4, 16bpp*2, 32bpp*1
#define SIZE 32
#include "refreshxy.c"

void Scr::Refresh48()		// 16bpp*3
#define SIZE 48
#include "refreshxy.c"

void Scr::Refresh64()		// 16bpp*4, 32bpp*2
#define SIZE 64
#include "refreshxy.c"

void Scr::Refresh96()		// 32bpp*3
#define SIZE 96
#include "refreshxy.c"

void Scr::Refresh128()		// 32bpp*4
#define SIZE 128
#include "refreshxy.c"


void Scr::Refresh24x1()		// 24bpp*1
#define SIZE -24
#include "refreshxy.c"

void Scr::Refresh24x2()		// 24bpp*2
#define SIZE -48
#include "refreshxy.c"

void Scr::Refresh24x3()		// 24bpp*3
#define SIZE -72
#include "refreshxy.c"

void Scr::Refresh24x4()		// 24bpp*4
#define SIZE -96
#include "refreshxy.c"


/* ----	draw horizontal line -----------------------------
	color is given as screenbyte ink
	x, y and pixelcount count are given normalized to scale=1
	caveat: line must start and end on byte boundary!
*/
void Scr::HLine ( long ink, int x, int y, int count )
{
	XXXLogLine(">Scr::HLine(%i,%i,%i,%i)",ink,x,y,count);

	if (window.Depth()==24)
	{
		long n  = count * scale;	// window pixels
		char *z = window.Data(x,y);	// start pointer
		
		while ( n>1 ) { *(long*)z=ink; z+=3; n--; }
		if (n>0) *(long*)z = (*(long*)z&0xff000000) + ink;
	}
	else
	{	
		long  n = count * scale * window.Depth() /8;	// bytes
		long *z = (long*)window.Data(x,y);		// start pointer
		long *e = (long*)((ptr)z+n);

		if (n&1) *((char*&)z)++  = (char)ink;
		if (n&2) *((short*&)z)++ = (short)ink;

		while (z<e) *z++ = ink;
	}

	window.Update(x,y,count);
	
	XXXLogExit();
}


/* ----	copy offscreen buffer to window ---------------------------
*/
void Scr::Sync ( )
{	long n;
	int i,j;
	long* dp;
	char* bp;

	XXXLogLine ( ">Scr::Sync()" );

// check flashing colors
//	flashing is done by:
//	1) invalidating attribute bytes which have the flash bit set
//	2) swapping colors in ink[] and pap[] in the flashing range

	n = frameCounter-nextflash;

	if (n>=0)
	{
		XXXLog("flashing");

		if ((Long)n>=32) nextflash=frameCounter;

		nextflash += 32;
		flashphase = !flashphase;

		if (flashphase)
		{	MemCopy(ink,ink+128,sizeof(*ink)*128);
			MemCopy(pap,pap+128,sizeof(*pap)*128);
		}
		else
		{	MemCopy(pap,ink+128,sizeof(*ink)*128);
			MemCopy(ink,pap+128,sizeof(*pap)*128);
		};

		XXXTRAP(NELEM(attr)!=SCR_LINES);

		Ptr a = attr[0];
		dp = dirty;
		for ( i=0; i<(int)NELEM(attr); i++ )
		{
			for ( j=32; j; j-- ) n += n + ( *a++>>7 );
			*dp++ |= n;
			XXXLog("%08X",n);
		}
	}

// check for screen size change
// then we have to recalculate some values
	if (window.Resized())
	{
		XXXLog("resized!");

		// new scale:
		width  = window.RealWidth();	// unscaled window width  in pixels
		height = window.RealHeight();	// unscaled window height in pixels

		scale = Min ( width/SCR_HSIZE, height/SCR_VSIZE );
		scale = Min ( SCR_MAXSCALE, scale );		// new magnification
		XTRAP(scale<1);
		window.SetScale(scale);
		width  = window.Width();	// width normalized to scale=1
		height = window.Height();	// height ""

		// new routine to use to refresh offscreen image
		char n[] = {1,2,3,4,6,8,12,16,24,32,48,64,96,128,0};
		method = window.Depth()==24 ? -scale : strchr(n,window.Depth()*scale)-n;

		// new location of screen within window:
		xa = (width - SCR_HSIZE) / 2;
		xe = xa + SCR_HSIZE;
		ya = (height - SCR_VSIZE) / 2;
		ye = ya + SCR_VSIZE;

		// set border beyond SCR_ABOVE and SCR_BELOW to black
		//for (i=0;      i<ya-SCR_ABOVE; i++) HLine(color[0],0,i,width);
		//for (i=ye+SCR_BELOW; i<height; i++) HLine(color[0],0,i,width);

		// invalidate whole window:
		MemSet(0xff,dirty,sizeof(dirty));
		for (i=0;i<(int)NELEM(border);i++) border[i] |= 0xf8;
		window.FullUpdate();
	}


// update border: 
// draw line to xim, update window and clear dirty flag
	XXXLog("border");
// line above top border and below bottom border
	if (border[0]           <0) for ( i=0;      i<ya-SCR_ABOVE; i++ ) HLine(color[border[0]&7],0,i,width);
	if (border[SCR_BORDER-1]<0) for ( i=ye+SCR_BELOW; i<height; i++ ) HLine(color[border[SCR_BORDER-1]&7],0,i,width);
// top border		
	for ( bp=border+SCR_ABOVE-LinesAbove(), i=ya-LinesAbove(); i<ya; bp++,i++ )
		if (*bp<0) { HLine(color[*bp&=7],0,i,width); }
// left and right of screen
	XXTRAP((xa>0)!=(width>xe));
	if (xa<=0) { bp+=ye-i;i=ye; }
	else for ( ; i<ye; bp++,i++)
		if (*bp<0)
		{	HLine(color[*bp&=7],0, i,xa);
			HLine(color[*bp   ],xe,i,width-xe);
		}
// bottom border
	for ( j=ye+LinesBelow(); i<j; bp++,i++)
		if (*bp<0) { HLine(color[*bp&=7],0,i,width); }


// update screen
	XXXLog("screen");
	switch(method)
	{
	case -4:Refresh24x4();	break;
	case -3:Refresh24x3();	break;
	case -2:Refresh24x2();	break;
	case -1:Refresh24x1();	break;
	case 0:	Refresh1();	break;
	case 1:	Refresh2();	break;
	case 2:	Refresh3();	break;
	case 3:	Refresh4();	break;
	case 4:	Refresh6();	break;
	case 5:	Refresh8();	break;
	case 6:	Refresh12();	break;
	case 7:	Refresh16();	break;
	case 8:	Refresh24();	break;
	case 9:	Refresh32();	break;
	case 10:Refresh48();	break;
	case 11:Refresh64();	break;
	case 12:Refresh96();	break;
	case 13:Refresh128();	break;
	default:IERR();
	}


// at last: flush buffers
	window.Sync();

	XXXLogExit();
}





