/*	Z80 Emulator: macro definitions, opcode enumerations

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	based on fMSX; Copyright (C) Marat Fayzullin 1994,1995
*/


// ----	definitions to ease the use of the z80 registers & core:

#define	IRPT		this->irpt		// irpt flags
#define	NMI		this->nmi		// nmi flag
#define	IFF1		this->iff1		// actually used irpt flip flop
#define	IFF2		this->iff2		// copy of iff1 during nmi processing
#define	IM		this->im			// interrupt mode: 0 ... 2
#define	RR		this->r			// 7 bit DRAM refresh counter
#define	RI		this->iic.r.hi		// hi byte of interrupt vector: i register
#define	IRPTCMD	this->iic.r.lo		// lo byte of interrupt vector: read from bus
#define	IRPTVEK	this->iic.rr		// interrupt vector in interrupt mode 2

#define	RA		this->af.r.hi
#define	RF		this->af.r.lo
#define	RA2		this->af2.r.hi
#define	RF2		this->af2.r.lo

#define	BC2		this->bc2.rr
#define	RB2		this->bc2.r.hi
#define	RC2		this->bc2.r.lo

#define	DE2		this->de2.rr
#define	RD2		this->de2.r.hi
#define	RE2		this->de2.r.lo

#define	HL2		this->hl2.rr
#define	RH2		this->hl2.r.hi
#define	RL2		this->hl2.r.lo

#define	BC		this->bc.rr
#define	RB		this->bc.r.hi
#define	RC		this->bc.r.lo

#define	DE		this->de.rr
#define	RD		this->de.r.hi
#define	RE		this->de.r.lo

#define	HL		this->hl.rr
#define	RH		this->hl.r.hi
#define	RL		this->hl.r.lo

#define	IX		this->ix.rr
#define	XH		this->ix.r.hi
#define	XL		this->ix.r.lo

#define	IY		this->iy.rr
#define	YH		this->iy.r.hi
#define	YL		this->iy.r.lo

#define	PC		this->pc.rr
#define	PCH		this->pc.r.hi
#define	PCL		this->pc.r.lo

#define	SP		this->sp.rr
#define	SPH		this->sp.r.hi
#define	SPL		this->sp.r.lo



// ----	bit masks for z80 flag register -------------------------------

#define S_FLAG	0x80
#define Z_FLAG	0x40
#define H_FLAG	0x10
#define P_FLAG	0x04
#define V_FLAG	0x04
#define N_FLAG	0x02
#define C_FLAG	0x01


// ----	read/write/modify memory --------------------------------------

#define	PEEK(A)	(rpage[(A)>>pagebits][A])			// read byte; addr must be const
#define	POKE(A,N)	(wpage[(A)>>pagebits][A]=N)		// write byte; addr must be const
#define	n		(wm=pc++,PEEK(wm))			// get next byte via pc
#define	nn		(pc+=2,PEEK(pc-2)+PEEK(pc-1)*256)	// get next word via pc
#define	dis		(char)n					// get signed char via pc
#define	nextopcode	(wm=pc++, mpage[wm>>pagebits][wm])	// M1 fetch

#define	POP()		(wm=sp++,PEEK(wm))
#define	PUSH(N)	(sp--,POKE(sp,N))
#define	POPD()		(c=POP(),c+POP()*256)
#define	PUSHD(N)	(PUSH(N>>8),PUSH(N))


// ----	Profiling of used opcodes and executed locations --------------------
//	depending on flags defined in z80.options

#if PC_PROFILE
    #define	COUNT_PC		{ if (count_pc) if (!++cnt_pc[pc]) Info_1st_Loc(cc,pc); }
#else
    #define	COUNT_PC		
#endif

#if CMD_PROFILE
    #define	COUNT_INSTR		{ if (count_instr && !++cnt_xx[PEEK(pc)] )  Info_1st_Instr(cc,pc); } 
    #define	COUNT_CB_INSTR	{ if (count_instr && !++cnt_cb[PEEK(pc)] )  Info_1st_Instr(cc,pc); }
    #define	COUNT_ED_INSTR	{ if (count_instr && !++cnt_ed[PEEK(pc)] )  Info_1st_Instr(cc,pc); }
    #define	COUNT_XY_INSTR	{ if (count_instr && !++cnt_xy[PEEK(pc)] )  Info_1st_Instr(cc,pc); }
    #define	COUNT_XYCB_INSTR	{ if (count_instr && !++cnt_xycb[PEEK(pc)]) Info_1st_Instr(cc,pc); }
#else
    #define	COUNT_INSTR		
    #define	COUNT_CB_INSTR	
    #define	COUNT_ED_INSTR	
    #define	COUNT_XY_INSTR	
    #define	COUNT_XYCB_INSTR	
#endif


// ----	Macros which call Z80_Info(cc,ip) or do nothing ---------------------------------
//	depending on flags defined in z80.options


#if INFO_IRPT
    #define	DO_INFO_IRPT		Info_Irpt ( cc, pc )
#else
    #define	DO_INFO_IRPT		
#endif


#if INFO_NMI
    #define	DO_INFO_NMI		Info_NMI ( cc, pc )
#else
    #define	DO_INFO_NMI		
#endif


#if INFO_ILLEGALS
    #define	INFO_ILL1		Info_Illegal ( cc, pc-1 )
    #define	INFO_ILL2		Info_Illegal ( cc, pc-2 )
    #define	INFO_ILL3		Info_Illegal ( cc, pc-3 )
    #define	INFO_ILL4		Info_Illegal ( cc, pc-4 )
#else
    #define	INFO_ILL1		
    #define	INFO_ILL2		
    #define	INFO_ILL3		
    #define	INFO_ILL4		
#endif


#if INFO_WEIRD
    #define	INFO_WEIRD2		Info_Weird ( cc, pc-2 )
    #define	INFO_WEIRD3		Info_Weird ( cc, pc-3 )
    #define	INFO_WEIRD4		Info_Weird ( cc, pc-4 )
#else
    #define	INFO_WEIRD2		 
    #define	INFO_WEIRD3		
    #define	INFO_WEIRD4		
#endif


#if INFO_MISC
    #define	INFO_MISC1		Info_Misc ( cc, pc-1 )
    #define	INFO_MISC2		Info_Misc ( cc, pc-2 )
#else
    #define	INFO_MISC1
    #define	INFO_MISC2
#endif


// ----	Exact register R handling:  R is kept in local variable
#if EXACT_R	
    #define	increment_r		r++
    #define	decrement_r		r--
    #define	add_r(N)		r += N
    #define	store_r			RR = (RR&0x80) + (r&0x7F)
    #define	load_r			r = RR
    #define	ld_r_a			RR = r = ra
    #define	ld_a_r			ra = (RR&0x80) + (r&0x7F)

// ----	Use register R for random generation only: Bit 7 is kept in RR, bits 0 ... 6 are 'guessed'
#else
    #define	increment_r	
    #define	decrement_r	
    #define	add_r(N)	
    #define	store_r		
    #define	load_r		
    #define	ld_r_a			RR = ra
    #define	ld_a_r			ra = (RR&0x80) + ((((Long)(TOTAL-CYCLES))/7)&0x7F)
#endif


// --------------------------------------------------------------------
// ----	INSTRUCTION MACROS --------------------------------------------
// --------------------------------------------------------------------

#define	OUTPUT(A,B)	myboard.Output(cc,A,B)
#define	INPUT(A)	myboard.Input(cc,A)


/*	RLC ... SRL:	set/clr C, Z, P, S;
			clear	N=0, H=0
			pres.	none
*/
#define M_RLC(R)      				\
	rf  = R>>7;				\
	R   = (R<<1)+rf; 				\
	rf |= zlog_flags[R]

#define M_RRC(R)      				\
	rf  = R&0x01;				\
	R   = (R>>1)+(rf<<7);		 	\
	rf |= zlog_flags[R]

#define M_RL(R)					\
	if (R&0x80)				\
	{	R 	= (R<<1)+(rf&0x01);	\
		rf	= zlog_flags[R]+C_FLAG;	\
	} else					\
	{	R 	= (R<<1)+(rf&0x01);	\
		rf	= zlog_flags[R];		\
	}

#define M_RR(R)					\
	if (R&0x01)				\
	{	R 	= (R>>1)+(rf<<7);		\
		rf	= zlog_flags[R]+C_FLAG;	\
	} else					\
	{	R 	= (R>>1)+(rf<<7);		\
		rf	= zlog_flags[R];		\
	}
  
#define M_SLA(R)				\
	rf	= R>>7;				\
	R <<= 1;					\
	rf |= zlog_flags[R]

#define M_SRA(R)				\
	rf	= R&0x01;			\
	R	= (R&0x80)+(R>>1);		\
	rf |= zlog_flags[R]

#define M_SLL(R)				\
	rf	= R>>7;				\
	R	= (R<<1)+1;			\
	rf |= zlog_flags[R]

#define M_SRL(R)				\
	rf	= R&0x01;			\
	R >>= 1;					\
	rf |= zlog_flags[R]


/*	BIT:	set/clr	Z
		clear	N=0, H=1
		pres	C
		takes other flags from corresponding bits in tested byte!
*/
#define M_BIT(N,R)	rf	= (rf&C_FLAG) + (R&(S_FLAG+P_FLAG)) + H_FLAG + ((R&N)?0:Z_FLAG)


/*	ADD ... CP:	set/clr	Z, S, V, C, N, H
				pres	none
*/
#define M_ADD(R)						\
	wm	= ra+R;						\
	rf	= wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			 + (~(ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			 + ((ra^R^wml)&H_FLAG);		\
	ra	= wml

#define M_SUB(R)						\
	wm	= ra-R;						\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;	\
	ra	= wml

#define M_ADC(R)						\
	wm	= ra+R+(rf&C_FLAG);				\
	rf	= wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			 + (~(ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			 + ((ra^R^wml)&H_FLAG);		\
	ra	= wml

#define M_SBC(R)						\
	wm	= ra-R-(rf&C_FLAG);				\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;	\
	ra	= wml

#define M_CP(R)						\
	wm	= ra-R;						\
	rf	= -wmh + (wml?0:Z_FLAG) + (wml&S_FLAG)	\
			  + ((ra^R)&(wml^ra)&0x80?V_FLAG:0)	\
			  + ((ra^R^wml)&H_FLAG) + N_FLAG;


/*	AND ... XOR:	set/clr	Z, P, S
					clear	C=0, N=0, H=0/1 (OR,XOR/AND)
					pres	none
*/
#define M_AND(R)			\
	ra &= R;				\
	rf	= H_FLAG|zlog_flags[ra]

#define M_OR(R)			\
	ra |= R;				\
	rf	= zlog_flags[ra]

#define M_XOR(R)			\
	ra ^= R;				\
	rf	= zlog_flags[ra]


/*	INC ... DEC:	set/clr	Z,P,S,H
					clear	N=0/1 (INC/DEC)
					pres	C
*/
#define M_INC(R)			\
	R++;				\
	rf	= (rf&C_FLAG) + (R?0:Z_FLAG) + (R&S_FLAG) + (R==0x80?V_FLAG:0) + (R&0x0F?0:H_FLAG)

#define M_DEC(R)			\
	R--;				\
	rf	= (rf&C_FLAG) + (R?0:Z_FLAG) + (R&S_FLAG) + (R==0x7F?V_FLAG:0) + (((R+1)&0x0F)?0:H_FLAG) + N_FLAG


/*	ADDW:	set/clr	C
			clear	N=0
			pres	Z, P, S
			unkn	H
*/
#define M_ADDW(R1,R2)					\
	rf &= ~(N_FLAG+C_FLAG);			\
	rf |= ((Long)R1+(Long)R2)>>16;			\
	R1 += R2;


/*	ADCW, SBCW:	set/clr	C,Z,V,S
			clear	N=0/1 (ADC/SBC)
			unkn	H
			pres	none
*/
#define M_ADCW(R)						\
	wm	= HL+R+(rf&C_FLAG);				\
	rf	= (((Long)HL+(Long)R+(rf&C_FLAG))>>16)		\
			+ (wm?0:Z_FLAG) + (wmh&S_FLAG)	\
			+ (~(HL^R)&(wm^HL)&0x8000?V_FLAG:0);\
	HL	= wm
   
#define M_SBCW(R)						\
	wm	= HL-R-(rf&C_FLAG);				\
	rf	= (((Long)HL-(Long)R-(rf&C_FLAG))>>31)	\
			+ (wm?0:Z_FLAG) + (wmh&S_FLAG)	\
			+ ((HL^R)&(wm^HL)&0x8000?V_FLAG:0)	\
			+ N_FLAG;				\
	HL	= wm

/*	IN	set/clr	Z, P, S, H
		clear	N=0
		pres	C
*/
#define M_IN(R)							\
	R	= INPUT(BC);					\
	rf	= (rf&C_FLAG) + zlog_flags[R]










